# 2024-12-22 AMH - Reloader tests have not been working for a while.
# We need to re-implement them.

# import os
# import secrets
# import sys

# from contextlib import suppress
# from subprocess import PIPE, Popen, TimeoutExpired
# from tempfile import TemporaryDirectory
# from textwrap import dedent
# from threading import Timer
# from time import sleep

# import pytest


# # We need to interrupt the autoreloader without killing it,
# # so that the server gets terminated
# # https://stefan.sofa-rockers.org/2013/08/15/handling-sub-process-hierarchies-python-linux-os-x/

# try:
#     from signal import CTRL_BREAK_EVENT
#     from subprocess import CREATE_NEW_PROCESS_GROUP

#     flags = CREATE_NEW_PROCESS_GROUP
# except ImportError:
#     flags = 0

# TIMER_DELAY = 2


# def terminate(proc):
#     if flags:
#         proc.send_signal(CTRL_BREAK_EVENT)
#     else:
#         proc.terminate()


# def write_app(filename, **runargs):
#     text = secrets.token_urlsafe()
#     with open(filename, "w") as f:
#         f.write(
#             dedent(
#                 f"""\
#             import os
#             from sanic import Sanic

#             app = Sanic(__name__)

#             app.route("/")(lambda x: x)

#             @app.listener("after_server_start")
#             def complete(*args):
#                 print("complete", os.getpid(), {text!r})

#             if __name__ == "__main__":
#                 app.run(**{runargs!r})
#             """
#             )
#         )
#     return text


# def write_listener_app(filename, **runargs):
#     start_text = secrets.token_urlsafe()
#     stop_text = secrets.token_urlsafe()
#     with open(filename, "w") as f:
#         f.write(
#             dedent(
#                 f"""\
#             import os
#             from sanic import Sanic

#             app = Sanic(__name__)

#             app.route("/")(lambda x: x)

#             @app.reload_process_start
#             async def reload_start(*_):
#                 print("reload_start", os.getpid(), {start_text!r})

#             @app.reload_process_stop
#             async def reload_stop(*_):
#                 print("reload_stop", os.getpid(), {stop_text!r})

#             if __name__ == "__main__":
#                 app.run(**{runargs!r})
#             """
#             )
#         )
#     return start_text, stop_text


# def write_json_config_app(filename, jsonfile, **runargs):
#     with open(filename, "w") as f:
#         f.write(
#             dedent(
#                 f"""\
#             import os
#             from sanic import Sanic
#             import json

#             app = Sanic(__name__)
#             with open("{jsonfile}", "r") as f:
#                 config = json.load(f)
#             app.config.update_config(config)

#             app.route("/")(lambda x: x)

#             @app.listener("after_server_start")
#             def complete(*args):
#                 print("complete", os.getpid(), app.config.FOO)

#             if __name__ == "__main__":
#                 app.run(**{runargs!r})
#             """
#             )
#         )


# def write_file(filename):
#     text = secrets.token_urlsafe()
#     with open(filename, "w") as f:
#         f.write(f"""{{"FOO": "{text}"}}""")
#     return text


# def scanner(proc, trigger="complete"):
#     for line in proc.stdout:
#         line = line.decode().strip()
#         if line.startswith(trigger):
#             yield line


# argv = dict(
#     script=[sys.executable, "reloader.py"],
#     module=[sys.executable, "-m", "reloader"],
#     sanic=[
#         sys.executable,
#         "-m",
#         "sanic",
#         "--port",
#         "42204",
#         "--auto-reload",
#         "reloader.app",
#     ],
# )


# @pytest.mark.parametrize(
#     "runargs, mode",
#     [
#         (dict(port=42202, auto_reload=True), "script"),
#         (dict(port=42203, auto_reload=True), "module"),
#         ({}, "sanic"),
#     ],
# )
# @pytest.mark.xfail
# async def test_reloader_live(runargs, mode):
#     with TemporaryDirectory() as tmpdir:
#         filename = os.path.join(tmpdir, "reloader.py")
#         text = write_app(filename, **runargs)
#         command = argv[mode]
#         proc = Popen(command, cwd=tmpdir, stdout=PIPE, creationflags=flags)
#         try:
#             timeout = Timer(TIMER_DELAY, terminate, [proc])
#             timeout.start()
#             # Python apparently keeps using the old source sometimes if
#             # we don't sleep before rewrite (pycache timestamp problem?)
#             sleep(1)
#             line = scanner(proc)
#             assert text in next(line)
#             # Edit source code and try again
#             text = write_app(filename, **runargs)
#             assert text in next(line)
#         finally:
#             timeout.cancel()
#             terminate(proc)
#             with suppress(TimeoutExpired):
#                 proc.wait(timeout=3)


# @pytest.mark.parametrize(
#     "runargs, mode",
#     [
#         (dict(port=42302, auto_reload=True), "script"),
#         (dict(port=42303, auto_reload=True), "module"),
#         ({}, "sanic"),
#     ],
# )
# @pytest.mark.xfail
# async def test_reloader_live_with_dir(runargs, mode):
#     with TemporaryDirectory() as tmpdir:
#         filename = os.path.join(tmpdir, "reloader.py")
#         config_file = os.path.join(tmpdir, "config.json")
#         runargs["reload_dir"] = tmpdir
#         write_json_config_app(filename, config_file, **runargs)
#         text = write_file(config_file)
#         command = argv[mode]
#         if mode == "sanic":
#             command += ["--reload-dir", tmpdir]
#         proc = Popen(command, cwd=tmpdir, stdout=PIPE, creationflags=flags)
#         try:
#             timeout = Timer(TIMER_DELAY, terminate, [proc])
#             timeout.start()
#             # Python apparently keeps using the old source sometimes if
#             # we don't sleep before rewrite (pycache timestamp problem?)
#             sleep(1)
#             line = scanner(proc)
#             assert text in next(line)
#             # Edit source code and try again
#             text = write_file(config_file)
#             assert text in next(line)
#         finally:
#             timeout.cancel()
#             terminate(proc)
#             with suppress(TimeoutExpired):
#                 proc.wait(timeout=3)


# def test_reload_listeners():
#     with TemporaryDirectory() as tmpdir:
#         filename = os.path.join(tmpdir, "reloader.py")
#         start_text, stop_text = write_listener_app(
#             filename, port=42305, auto_reload=True
#         )

#         proc = Popen(
#             argv["script"], cwd=tmpdir, stdout=PIPE, creationflags=flags
#         )
#         try:
#             timeout = Timer(TIMER_DELAY, terminate, [proc])
#             timeout.start()
#             # Python apparently keeps using the old source sometimes if
#             # we don't sleep before rewrite (pycache timestamp problem?)
#             sleep(1)
#             line = scanner(proc, "reload_start")
#             assert start_text in next(line)
#             line = scanner(proc, "reload_stop")
#             assert stop_text in next(line)
#         finally:
#             timeout.cancel()
#             terminate(proc)
#             with suppress(TimeoutExpired):
#                 proc.wait(timeout=3)



from multiprocessing.connection import Connection
from os import environ, getpid
from typing import Any, Callable, Optional

from sanic.log import Colors, logger
from sanic.worker.process import ProcessState
from sanic.worker.state import WorkerState


class WorkerMultiplexer:
    """Multiplexer for Sanic workers.

    This is instantiated inside of worker porocesses only. It is used to
    communicate with the monitor process.

    Args:
        monitor_publisher (Connection): The connection to the monitor.
        worker_state (Dict[str, Any]): The state of the worker.
    """

    def __init__(
        self,
        monitor_publisher: Connection,
        worker_state: dict[str, Any],
    ):
        self._monitor_publisher = monitor_publisher
        self._state = WorkerState(worker_state, self.name)

    def ack(self):
        """Acknowledge the worker is ready."""
        logger.debug(
            f"{Colors.BLUE}Process ack: {Colors.BOLD}{Colors.SANIC}"
            f"%s {Colors.BLUE}[%s]{Colors.END}",
            self.name,
            self.pid,
        )
        self._state._state[self.name] = {
            **self._state._state[self.name],
            "state": ProcessState.ACKED.name,
        }

    def manage(
        self,
        ident: str,
        func: Callable[..., Any],
        kwargs: dict[str, Any],
        transient: bool = False,
        restartable: Optional[bool] = None,
        tracked: bool = False,
        auto_start: bool = True,
        workers: int = 1,
    ) -> None:
        """Manages the initiation and monitoring of a worker process.

        Args:
            ident (str): A unique identifier for the worker process.
            func (Callable[..., Any]): The function to be executed in the worker process.
            kwargs (Dict[str, Any]): A dictionary of arguments to be passed to `func`.
            transient (bool, optional): Flag to mark the process as transient. If `True`,
                the Worker Manager will restart the process with any global restart
                (e.g., auto-reload). Defaults to `False`.
            restartable (Optional[bool], optional): Flag to mark the process as restartable. If `True`,
                the Worker Manager can restart the process if prompted. Defaults to `None`.
            tracked (bool, optional): Flag to indicate whether the process should be tracked
                after its completion. Defaults to `False`.
            auto_start (bool, optional): Flag to indicate whether the process should be started
            workers (int, optional): The number of worker processes to run. Defaults to 1.

        This method packages the provided arguments into a bundle and sends them back to the
        main process to be managed by the Worker Manager.
        """  # noqa: E501
        bundle = (
            ident,
            func,
            kwargs,
            transient,
            restartable,
            tracked,
            auto_start,
            workers,
        )
        self._monitor_publisher.send(bundle)

    def set_serving(self, serving: bool) -> None:
        """Set the worker to serving.

        Args:
            serving (bool): Whether the worker is serving.
        """
        self._state._state[self.name] = {
            **self._state._state[self.name],
            "serving": serving,
        }

    def exit(self):
        """Run cleanup at worker exit."""
        try:
            del self._state._state[self.name]
        except ConnectionRefusedError:
            logger.debug("Monitor process has already exited.")

    def restart(
        self,
        name: str = "",
        all_workers: bool = False,
        zero_downtime: bool = False,
    ):
        """Restart the worker.

        Args:
            name (str): The name of the process to restart.
            all_workers (bool): Whether to restart all workers.
            zero_downtime (bool): Whether to restart with zero downtime.
        """
        if name and all_workers:
            raise ValueError(
                "Ambiguous restart with both a named process and"
                " all_workers=True"
            )
        if not name:
            name = "__ALL_PROCESSES__:" if all_workers else self.name
        if not name.endswith(":"):
            name += ":"
        if zero_downtime:
            name += ":STARTUP_FIRST"
        self._monitor_publisher.send(name)

    reload = restart  # no cov
    """Alias for restart."""

    def scale(self, num_workers: int):
        """Scale the number of workers.

        Args:
            num_workers (int): The number of workers to scale to.
        """
        message = f"__SCALE__:{num_workers}"
        self._monitor_publisher.send(message)

    def terminate(self, early: bool = False):
        """Terminate the worker.

        Args:
            early (bool): Whether to terminate early.
        """
        message = "__TERMINATE_EARLY__" if early else "__TERMINATE__"
        self._monitor_publisher.send(message)

    @property
    def pid(self) -> int:
        """The process ID of the worker."""
        return getpid()

    @property
    def name(self) -> str:
        """The name of the worker."""
        return environ.get("SANIC_WORKER_NAME", "")

    @property
    def state(self):
        """The state of the worker."""
        return self._state

    @property
    def workers(self) -> dict[str, Any]:
        """The state of all workers."""
        return self.state.full()



from __future__ import annotations

from asyncio import BaseTransport
from typing import TYPE_CHECKING, Optional, Union


if TYPE_CHECKING:
    from sanic.http.constants import HTTP
    from sanic.models.asgi import ASGIScope


# Protocol is a 3.8+ feature
from typing import Protocol


class HTMLProtocol(Protocol):
    def __html__(self) -> Union[str, bytes]: ...

    def _repr_html_(self) -> Union[str, bytes]: ...


class Range(Protocol):
    start: Optional[int]
    end: Optional[int]
    size: Optional[int]
    total: Optional[int]
    __slots__ = ()


class TransportProtocol(BaseTransport):
    scope: ASGIScope
    version: HTTP
    __slots__ = ()



import sys

from multiprocessing import Event
from os import environ, getpid
from typing import Any, Union
from unittest.mock import Mock

import pytest

from sanic import Sanic
from sanic.compat import use_context
from sanic.worker.multiplexer import WorkerMultiplexer
from sanic.worker.state import WorkerState


def noop(*args, **kwargs):
    pass


@pytest.fixture
def monitor_publisher():
    return Mock()


@pytest.fixture
def worker_state():
    return {}


@pytest.fixture
def m(monitor_publisher, worker_state):
    environ["SANIC_WORKER_NAME"] = "Test"
    worker_state["Test"] = {}
    yield WorkerMultiplexer(monitor_publisher, worker_state)
    del environ["SANIC_WORKER_NAME"]


@pytest.mark.skipif(
    sys.platform not in ("linux", "darwin"),
    reason="This test requires fork context",
)
def test_has_multiplexer_default(app: Sanic):
    event = Event()

    @app.main_process_start
    async def setup(app, _):
        app.shared_ctx.event = event

    @app.after_server_start
    def stop(app):
        if hasattr(app, "m") and isinstance(app.m, WorkerMultiplexer):
            app.shared_ctx.event.set()
        app.stop()

    with use_context("fork"):
        app.run()

    assert event.is_set()


def test_not_have_multiplexer_single(app: Sanic):
    event = Event()

    @app.main_process_start
    async def setup(app, _):
        app.shared_ctx.event = event

    @app.after_server_start
    def stop(app):
        if hasattr(app, "m") and isinstance(app.m, WorkerMultiplexer):
            app.shared_ctx.event.set()
        app.stop()

    app.run(single_process=True)

    assert not event.is_set()


def test_ack(worker_state: dict[str, Any], m: WorkerMultiplexer):
    worker_state["Test"] = {"foo": "bar"}
    m.ack()
    assert worker_state["Test"] == {"foo": "bar", "state": "ACKED"}


def test_restart_self(monitor_publisher: Mock, m: WorkerMultiplexer):
    m.restart()
    monitor_publisher.send.assert_called_once_with("Test:")


def test_restart_foo(monitor_publisher: Mock, m: WorkerMultiplexer):
    m.restart("foo")
    monitor_publisher.send.assert_called_once_with("foo:")


def test_reload_alias(monitor_publisher: Mock, m: WorkerMultiplexer):
    m.reload()
    monitor_publisher.send.assert_called_once_with("Test:")


def test_terminate(monitor_publisher: Mock, m: WorkerMultiplexer):
    m.terminate()
    monitor_publisher.send.assert_called_once_with("__TERMINATE__")


def test_scale(monitor_publisher: Mock, m: WorkerMultiplexer):
    m.scale(99)
    monitor_publisher.send.assert_called_once_with("__SCALE__:99")


def test_manage(monitor_publisher: Mock, m: WorkerMultiplexer):
    m.manage("NEW", noop, auto_start=False, kwargs={"foo": 99})
    monitor_publisher.send.assert_called_once_with(
        ("NEW", noop, {"foo": 99}, False, None, False, False, 1)
    )


def test_properties(
    monitor_publisher: Mock, worker_state: dict[str, Any], m: WorkerMultiplexer
):
    assert m.reload == m.restart
    assert m.pid == getpid()
    assert m.name == "Test"
    assert m.workers == worker_state
    assert m.state == worker_state["Test"]
    assert isinstance(m.state, WorkerState)


@pytest.mark.parametrize(
    "params,expected",
    (
        ({}, "Test:"),
        ({"name": "foo"}, "foo:"),
        ({"all_workers": True}, "__ALL_PROCESSES__:"),
        ({"zero_downtime": True}, "Test::STARTUP_FIRST"),
        ({"name": "foo", "all_workers": True}, ValueError),
        ({"name": "foo", "zero_downtime": True}, "foo::STARTUP_FIRST"),
        (
            {"all_workers": True, "zero_downtime": True},
            "__ALL_PROCESSES__::STARTUP_FIRST",
        ),
        (
            {"name": "foo", "all_workers": True, "zero_downtime": True},
            ValueError,
        ),
    ),
)
def test_restart_params(
    monitor_publisher: Mock,
    m: WorkerMultiplexer,
    params: dict[str, Any],
    expected: Union[str, type[Exception]],
):
    if isinstance(expected, str):
        m.restart(**params)
        monitor_publisher.send.assert_called_once_with(expected)
    else:
        with pytest.raises(expected):
            m.restart(**params)
